#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include "readmatrix.h"

main(int argc, char *argv[])
{
	const int tag = 50;
	int id, ntasks, source_id, dest_id, rc, i,j;
	MPI_Status status;
	int msg[2];

	float **arr;
	int result=0;
	char *filename;
	int size,rowsize;
	float threshold,max,curmax;

	
	int start,stop;

	rc = MPI_Init(&argc, &argv);


	if (rc != MPI_SUCCESS) {

		printf("MPI Init failed\n");
		exit(1);
	}
	rc = MPI_Comm_size(MPI_COMM_WORLD, &ntasks);
	rc = MPI_Comm_rank(MPI_COMM_WORLD, &id);

						
	if ( id != 0 ) { 
		float up, down, left, right;
		float old, change, maxchange;
		int cnt;
	
		dest_id = 0;
		rc = MPI_Recv(&size, 1, MPI_INT, dest_id, tag, MPI_COMM_WORLD, &status);
		rc = MPI_Recv(&rowsize, 1, MPI_INT, dest_id, tag, MPI_COMM_WORLD, &status);
		arr = malloc(sizeof(int *)*size);
		for (i=0;i<size; i++) {
			arr[i] = malloc(sizeof(int) * rowsize);
			rc = MPI_Recv(arr[i], rowsize, MPI_FLOAT, dest_id, tag, MPI_COMM_WORLD, &status);
		}
		fprintf(stderr,"task %i received data\n",id);
		for (i=0;i<size; i++) {
			for (j=0;j<rowsize;j++) {
				printf("%f ", arr[i][j]);
			}
			printf("\n");
		}
		
		do {
			
		maxchange =0;	
		for (i=1;i<size -1; i++) {
			start = 0;
			stop = rowsize;
			if (id == 1)
				start++;
			if (id == ntasks -1)
				stop--;
			for (j=start;j<stop;j++) {
				if (j == 0) {
					fprintf(stderr,"task %i is about to send data to the left, row %i\n",id,i);
					rc = MPI_Send(&(arr[i][0]), 1, MPI_FLOAT, id-1, tag, MPI_COMM_WORLD);
					fprintf(stderr,"task %i is about to receive data from the left, row %i\n",id,i);
					rc = MPI_Recv(&left, 1, MPI_FLOAT, id -1, tag, MPI_COMM_WORLD,&status);
					fprintf(stderr,"task %i has received data from the left, row %i\n",id,i);
				} else 
					left = arr[i][j-1];
				up = arr[i-1][j];
				down = arr[i+1][j];
				if (j<rowsize -1)
					right = arr[i][j+1];
				else {
					fprintf(stderr,"task %i is about to receive data from the right, row %i\n",id,i);
					rc = MPI_Recv(&right,  1, MPI_FLOAT, id +1, tag, MPI_COMM_WORLD,&status);
					fprintf(stderr,"task %i has received data from the right, row %i\n",id,i);
				}
				old = arr[i][j];
				arr[i][j] = (left + right + up + down) / 4;
				change = old - arr[i][j];
				if (change > maxchange)
					maxchange = change;
				

				
			}
			if ( id != ntasks -1) {
				fprintf(stderr,"task %i is about to send data to the right, row %i\n",id,i);
				rc = MPI_Send(&(arr[i][rowsize-1]),1, MPI_FLOAT, id +1, tag, MPI_COMM_WORLD);
				fprintf(stderr,"task %i has sent data to the right, row %i\n",id,i);
			}
		}
		
		fprintf(stderr,"task %i sending maxch\n",id);
		rc = MPI_Send(&maxchange,1, MPI_FLOAT, 0, tag, MPI_COMM_WORLD);
		rc = MPI_Recv(&cnt,  1, MPI_INT, 0, tag, MPI_COMM_WORLD,&status);
		if (cnt == 1)
			break;

		} while (1);

	        for (i=0;i<size; i++) 
	                rc = MPI_Send(arr[i], rowsize, MPI_FLOAT, 0, tag, MPI_COMM_WORLD);

		
	} else { 
		int res,start,sublength;

		filename = argv[1];
	       readmatrix(filename, &arr, &size, &threshold);

		for (i=0;i < ntasks -1; i++) {

			start = size/(ntasks-1)*i;
			sublength = size/(ntasks-1);
			if (i == ntasks - 2 )
				sublength += size % (ntasks-1);
			printf("%i %i %i\n",i+1,start,sublength);
			rc = MPI_Send(&size, 1, MPI_INT, i+1, tag, MPI_COMM_WORLD);
			rc = MPI_Send(&sublength, 1, MPI_INT, i+1, tag, MPI_COMM_WORLD);
			for (j=0; j < size; j++) {
				rc = MPI_Send(&(arr[j][start]), sublength, MPI_FLOAT, i+1, tag, MPI_COMM_WORLD);
			}
	
	
		}
		
		do {
			int res;
			
			max = 0;
			printf("waiting for maxchanges\n");
			for (i=0;i < ntasks -1; i++) {
				rc = MPI_Recv(&curmax, 1, MPI_FLOAT, i+1, tag, MPI_COMM_WORLD,&status);
				if (curmax > max) 
					max = curmax;
			}
			if (max < threshold) {
				res = 1;
				for (i=0;i < ntasks -1; i++)
					rc = MPI_Send(&res, 1, MPI_INT, i+1, tag, MPI_COMM_WORLD);
				break;
			} else {
				res = 0;
				for (i=0;i < ntasks -1; i++)
					rc =  MPI_Send(&res, 1, MPI_INT, i+1, tag, MPI_COMM_WORLD);
			}
		} while (1);
				
		
		printf("waiting for result\n");

		for (i=0;i < ntasks -1; i++) {

		        start = size/(ntasks-1)*i;
		        sublength = size/(ntasks-1);
		        if (i == ntasks - 2 )
		                sublength += size % (ntasks-1);
		        for (j=0; j < size; j++) 
		                rc = MPI_Recv(&(arr[j][start]), sublength, MPI_FLOAT, i+1, tag, MPI_COMM_WORLD,&status);
	        }
		
		for (i=0;i < size; i++) {
			for (j=0; j< size; j++)
				printf("%3.2f ",arr[i][j]);
			printf("\n");
		}	
	
	}



	MPI_Finalize();

	exit(0);

}

