#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include "readmatrix.h"

main(int argc, char *argv[])
{
	const int tag = 50;
	int id, ntasks, source_id, dest_id, rc, i,j;
	MPI_Status status;
	int msg[2];

	int **arr;
	int result=0;
	char *filename;
	int size,rowsize;

	rc = MPI_Init(&argc, &argv);


	if (rc != MPI_SUCCESS) {

		printf("MPI Init failed\n");
		exit(1);
	}
	rc = MPI_Comm_size(MPI_COMM_WORLD, &ntasks);
	rc = MPI_Comm_rank(MPI_COMM_WORLD, &id);

						
	if ( id != 0 ) { 
		int up, down, left, right;
	
		dest_id = 0;
		rc = MPI_Recv(&size, 1, MPI_INT, dest_id, tag, MPI_COMM_WORLD, &status);
		rc = MPI_Recv(&rowsize, 1, MPI_INT, dest_id, tag, MPI_COMM_WORLD, &status);
		arr = malloc(sizeof(int *)*size);
		for (i=0;i<size; i++) {
			arr[i] = malloc(sizeof(int) * rowsize);
			rc = MPI_Recv(arr[i], rowsize, MPI_INT, dest_id, tag, MPI_COMM_WORLD, &status);
		}
		for (i=0;i<size; i++) {
			for (j=0;j<rowsize;j++) {
				left = right = up = down = 0;
				if (id != 1 && j ==0) {
					rc = MPI_Send(&(arr[i][0]), 1, MPI_INT, id-1, tag, MPI_COMM_WORLD);
					rc = MPI_Recv(&left, 1, MPI_INT, id -1, tag, MPI_COMM_WORLD,&status);
				}
				if (j !=0)
					left = arr[i][j-1];
				if (i>0)
					up = arr[i-1][j];
				if (i<size-1)
					down = arr[i+1][j];
				if (j<rowsize -1)
					right = arr[i][j+1];
				else if (id != ntasks -1)
					rc = MPI_Recv(&right,  1, MPI_INT, id +1, tag, MPI_COMM_WORLD,&status);
				arr[i][j] = (left + right + up + down) / 4;
				if ( j == rowsize -1 && id != ntasks -1)
					rc = MPI_Send(&(arr[i][j]),1, MPI_INT, id +1, tag, MPI_COMM_WORLD);
			}
		}

	        for (i=0;i<size; i++) 
	                rc = MPI_Send(arr[i], rowsize, MPI_INT, 0, tag, MPI_COMM_WORLD);
	} else { 
		int res,start,sublength;

		filename = argv[1];
	       readmatrix(filename, &arr, &size);

		for (i=0;i < ntasks -1; i++) {

			start = size/(ntasks-1)*i;
			sublength = size/(ntasks-1);
			if (i == ntasks - 2 )
				sublength += size % (ntasks-1);
			printf("%i %i %i\n",i+1,start,sublength);
			rc = MPI_Send(&size, 1, MPI_INT, i+1, tag, MPI_COMM_WORLD);
			rc = MPI_Send(&sublength, 1, MPI_INT, i+1, tag, MPI_COMM_WORLD);
			for (j=0; j < size; j++) {
				rc = MPI_Send(&(arr[j][start]), sublength, MPI_INT, i+1, tag, MPI_COMM_WORLD);
			}
	
	
		}
		printf("waiting for result\n");

		for (i=0;i < ntasks -1; i++) {

		        start = size/(ntasks-1)*i;
		        sublength = size/(ntasks-1);
		        if (i == ntasks - 2 )
		                sublength += size % (ntasks-1);
		        for (j=0; j < size; j++) 
		                rc = MPI_Recv(&(arr[j][start]), sublength, MPI_INT, i+1, tag, MPI_COMM_WORLD,&status);
	        }
		
		for (i=0;i < size; i++) {
			for (j=0; j< size; j++)
				printf("%i ",arr[i][j]);
			printf("\n");
		}	
	
	}



	MPI_Finalize();

	exit(0);

}

