#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>

#define NUM_MSG 10
#define SHUTDOWN_MSG -1

const int tag = 50;


int getuplink(int id) 
{
/* we need to set the leftmost nonzero bit to zero */

	int i;
	
	for (i=1; i <= sizeof(int) * 256; i = i*2) {
		if ((id & i) != 0)
			return (id & (~i));
	}
	return 0;
}

int route(int msg, int id, int dest)
{
	int tmp,i,rc, route;
	
	tmp = id ^ dest;
	i=1;
	while (tmp != 1) {
		i = i * 2;
		tmp = tmp >> 1;
	}
	route = id | i;
	printf("task %i sending message %i to task %i via task %i\n",id,msg,dest,route);
	rc = MPI_Send(&msg,1,MPI_INT, route, tag,MPI_COMM_WORLD);
	rc = MPI_Send(&dest,1,MPI_INT, route, tag, MPI_COMM_WORLD);
	
}	

int setbits(int num)
{
	int bits = 0;
	while (num!=0) {
		if ((num & 1) == 1)
			bits++;
		num = num >> 1;
	}
	return bits;
}

int dim(int ntasks)
{
	int dim = 0;
	while (ntasks != 1) {
		ntasks = ntasks / 2;
		dim++;
	}
	return dim;
}

main(int argc, char *argv[])
{
	int id, ntasks, dest, rc, i,j;
	MPI_Status status;

	rc = MPI_Init(&argc, &argv);


	if (rc != MPI_SUCCESS) {

		printf("MPI Init failed\n");
		exit(1);
	}
	rc = MPI_Comm_size(MPI_COMM_WORLD, &ntasks);
	rc = MPI_Comm_rank(MPI_COMM_WORLD, &id);

						
	if ( id == 0 ) { 
		for (i=0; i < NUM_MSG; i++) {
			printf("task 0: message %i\n",i);
			
			dest = rand() % (ntasks-1) + 1;
			printf("task 0: message %i to task %i\n",i, dest);
			route(i, id, dest);
		}
		for (i = dim(ntasks); i > 0; i--) {
			for (j=1; j<ntasks;j++) {
				if (setbits(j) == i) {
					printf("task 0  sending shutdown message to task %i\n",j);
		                        route(SHUTDOWN_MSG, id , j);
				}
			}
		}
		
	} else {
		int uplink,msg;

		do {
			printf("task %i\n", id);
			uplink = getuplink(id);
			printf("task %i, uplink %i\n",id, uplink);
			rc = MPI_Recv(&msg, 1, MPI_INT, uplink, tag, MPI_COMM_WORLD,&status);
			rc = MPI_Recv(&dest, 1, MPI_INT, uplink, tag, MPI_COMM_WORLD,&status);
			if (dest!=id)
				route(msg, id, dest);
			else {
				if (msg == SHUTDOWN_MSG) {
					printf("task %i received shutdown message\n", id);
					break;
				} else {
					printf("task %i received msg %i\n", id, msg);
				}
			}
				
		} while (1);
				
		
	}


	MPI_Finalize();

	exit(0);

}

